import ast

from typing import Any
from functools import partial
from types import FunctionType

from .get_source_code import get_source_code

__all__ = ['get_real_func_name']


def get_real_func_name(fun: FunctionType or partial) -> str:
    """
    Parses a function and returns its 'real' name from its code.

    Args:
        fun (FunctionType): A target function.

    Returns:
        str: Function's parsed name.

    >>> def func_1(*args):
    ...     pass
    ...
    >>> get_real_func_name(func_1)
    'func_1'

    >>> def parent_func(*args, **kwargs):
    ...     def nested_func():
    ...         pass
    ...     return nested_func
    ...
    >>> get_real_func_name(parent_func)
    'parent_func'

    >>> get_real_func_name(parent_func())
    'nested_func'

    >>> get_real_func_name(partial(parent_func, func_1))
    'parent_func'
    """

    fun_code = get_source_code(fun)
    tree = ast.parse(fun_code)
    parse_func = _ParseFuncName()
    try:
        parse_func.visit(tree)
    except _FuncParseFinished:
        pass

    return parse_func.name


class _FuncParseFinished(StopIteration):
    """_ParseFuncName finished parsing."""

class _ParseFuncName(ast.NodeVisitor):
    """
    A node visitor that stores a 'real' name of a parsed function to 'name' attribute
    and raises FuncNameParsed to stop .
    """
    def __init__(self):
        self.name = None

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        self.name = node.name
        raise _FuncParseFinished



def jupyter_code(func_code: str):
    from IPython.display import Code
    return Code(func_code)


def display_code(func_code: str):
    from IPython.display import display
    display(jupyter_code(func_code))

__all__ = ['jupyter_code', 'display_code']



import logging
from pathlib import Path
from datetime import datetime as dt
from functools import wraps
from typing import List, Dict, Tuple, Any, Optional, Union
from types import FunctionType
from contextlib import contextmanager

import matplotlib.pyplot as plt

from .save_opts import SaveOpts, PLT_FORMAT, add_opt_docs
from ..file import save_plt_file, read_plt_file, PltFile


@add_opt_docs
class PltProject(object):

    """
    Args:
        folder (Path/str): A base folder of a new project. All the files will be saved to this folder.
        save_opts (SaveOpts, optional): Default save options. Can be provided as an object or using the keyword 
            arguments from below. 
        {save_opt_args}
    """

    _TEST: bool = False

    def __init__(self,
                 folder: Path or str,
                 save_opts: Optional[SaveOpts] = None,
                 **save_opt_kwargs,
                 ):
        self.folder = Path(folder)
        self.folder.mkdir(exist_ok=True)

        self.default_opts: SaveOpts = save_opts or SaveOpts(**save_opt_kwargs)
        self._context_opts: SaveOpts = SaveOpts()

    @property
    def context_opts(self) -> Optional[SaveOpts]:
        return self._context_opts

    @add_opt_docs
    def get_save_opts(self, **kwargs) -> SaveOpts:
        """
        Args:
            {save_opt_args}
        """
        return self.context_opts.\
            new_opts(only_missing_keys=True, **kwargs).\
            new_opts(only_missing_keys=True, **self.default_opts)

    @contextmanager
    @add_opt_docs
    def __call__(self,
                 save_opts: SaveOpts = None,
                 **save_opt_kwargs
                 ):
        """
        Args:
            folder (Path/str): A base folder of a new project. All the files will be saved to this folder.
            save_opts (SaveOpts, optional): Default save options. Can be provided as an object or using the keyword 
                arguments from below. 
        {save_opt_args}
        """

        self._context_opts = save_opts or SaveOpts(**save_opt_kwargs)
        try:
            yield
        finally:
            self._context_opts = SaveOpts()

    @add_opt_docs
    def save(self,
             func: FunctionType,
             args: Optional[Tuple[Any, ...]] = (),
             kwargs: Optional[Dict[str, Any]] = None,
             save_opts: Optional[SaveOpts] = None,
             **save_opt_kwargs,
             ) -> Path:

        """
        Save a function and its arguments based on provided save options.
        
        Args:
            func (FunctionType): Function to save.
            args (tuple, optional): Positional arguments.
            kwargs (dict, optional): Keyword arguments.
            save_opts (SaveOpts, optional): Current save options. Can be provided as an object or 
                using the keyword arguments from below.
            {save_opt_args}
        """
        save_opts = self.get_save_opts(**(save_opts or save_opt_kwargs))
        plt_name, fig_name = save_opts.get_names(func)
        plt_path = self.folder / plt_name

        kwargs = kwargs or {}

        if save_opts.save_figure:
            _save_figure(func, self.folder / fig_name, args, kwargs)

        if save_opts.save_plt:
            save_plt_file(plt_path, func, args=args, kwargs=kwargs, info_dict=save_opts.info_dict)

        return plt_path

    def register(self, func: FunctionType):
        @wraps(func)
        def wrapper(*args, **kwargs):
            self._register_called(func, args, kwargs)

        return wrapper

    @add_opt_docs
    def auto_save(self, save_opts: Optional[SaveOpts] = None, **save_opt_kwargs):
        """
        A decorator that registers a decorated function with
        Args:
            save_opts:
            {save_opt_args}

        Returns:

        """
        default_save_opts: SaveOpts = save_opts or SaveOpts(**save_opt_kwargs)

        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                if not self._TEST:
                    func(*args, **kwargs)
                self.save(func, args, kwargs, **default_save_opts)

            return wrapper

        return decorator

    def list_files(self) -> List[str]:
        return [path.stem for path in self.folder.glob(f'*.{PLT_FORMAT}')]

    def load_file(self, name: str) -> PltFile:
        return read_plt_file(self.folder / f'{name}.{PLT_FORMAT}')

    def __getitem__(self, name: str):
        return self.load_file(name)

    def _register_called(self, func: FunctionType, args: List[Any], kwargs: Dict[str, Any]):
        if not self._TEST:
            func(*args, **kwargs)
        if self._context_opts:
            self.save(func, args, kwargs)


def _save_figure(
        func: FunctionType,
        path: Path,
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any]
):

    path = str(path.absolute())
    plt.ioff()
    try:
        func(*args, **kwargs)
    except Exception as err:
        logger = logging.getLogger(__name__)
        logger.exception(err)
        logger.error(f'Function execution failed. Could not save to {path}.')
    plt.savefig(path)
    plt.ion()



from inspect import getsource
from functools import partial
from types import FunctionType


def get_source_code(func: FunctionType or partial) -> str:
    """
    Parse source code with fixed indentations, omitted decorators and partial.

    Args:
        func (FunctionType): A target function to parse.

    Returns:
        str: Source code with fixed indentations and omitted decorators.

    Examples:
        >>> def decorator(fun):
        ...     return fun
        ...
        >>> print(get_source_code(decorator).rstrip())
        def decorator(fun):
            return fun

        >>> def parent_func():
        ...     def nested_func(*args, **kwargs):
        ...         pass
        ...     return nested_func
        ...
        >>> print(get_source_code(parent_func).rstrip())
        def parent_func():
            def nested_func(*args, **kwargs):
                pass
            return nested_func

        >>> print(get_source_code(parent_func()).rstrip())
        def nested_func(*args, **kwargs):
            pass

        >>> @decorator
        ... def func():
        ...     pass
        ...
        >>> print(get_source_code(func).rstrip())
        def func():
            pass

        >>> print(get_source_code(partial(decorator, func)).rstrip())
        def decorator(fun):
            return fun



    """
    if isinstance(func, partial):
        func = func.func
    src = getsource(func)
    src = _fix_func_code(src)
    return src


def _fix_func_code(func_code: str):
    func_code = _fix_indents(func_code, ' ')
    func_code = _fix_indents(func_code, '\t')
    func_code = _remove_decorators(func_code)

    return func_code


def _remove_decorators(func_code: str):
    counts = 0
    lines = func_code.split('\n')

    for line in lines:
        if line.startswith('def '):
            break
        counts += 1
    return '\n'.join(lines[counts:])


def _fix_indents(func_code: str, char: str):
    counts: int = 0
    lines = func_code.split('\n')

    for ch in lines[0].split(char):
        if ch:
            break
        counts += 1
    return '\n'.join([line[counts:] for line in lines])



import ast
import builtins
from typing import List, Set, Dict, Any

__all__ = ['CheckUsedGlobals']


class CheckUsedGlobals(ast.NodeVisitor):
    builtins = dir(builtins)

    def __init__(self):
        self.modules: Set[str] = set()
        self.imported: Set[str] = set()
        self.args: Set[str] = set()
        self.assigned_names: Set[str] = set()
        self.used_globals: Set[str] = set()

    def _check_assigned(self, name: str):
        return (
                name in self.assigned_names or
                name in self.builtins or
                name in self.args
        )

    def clear(self):
        self.modules.clear()
        self.imported.clear()
        self.args.clear()
        self.assigned_names.clear()
        self.used_globals.clear()

    def visit_Import(self, node: ast.Import):
        for import_name in node.names:
            self.modules.add(import_name.name.split('.')[0])
            self.imported.add(import_name.name)

            if import_name.asname:
                self.assigned_names.add(import_name.asname)
            else:
                name = import_name.name.split('.')[-1]
                self.assigned_names.add(name)

        self.generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom):
        self.modules.add(node.module.split('.')[0])

        for import_name in node.names:
            self.imported.add('.'.join([node.module, import_name.name]))
            if import_name.asname:
                self.assigned_names.add(import_name.asname)
            else:
                self.assigned_names.add(import_name.name)

        self.generic_visit(node)

    def visit_Call(self, node: ast.Call) -> Any:
        for name in self._get_names(node):
            self.update_globals(name)
        self.generic_visit(node)

    def visit_For(self, node: ast.For) -> Any:
        self.assigned_names.update(list(self._get_names(node.target)))
        self.generic_visit(node)

    def visit_Assign(self, node: ast.Assign):
        for target_node in node.targets:
            self.assigned_names.update(list(self._get_names(target_node)))
        self.generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef):
        self.assigned_names.add(node.name)
        self.args.update(
            _get_funcdef_arg_names(node, 'posonlyargs') +
            _get_funcdef_arg_names(node, 'args') +
            _get_funcdef_arg_names(node, 'kwonlyargs')
        )
        if node.args.vararg:
            self.args.add(node.args.vararg.arg)
        if node.args.kwarg:
            self.args.add(node.args.kwarg.arg)
        self.generic_visit(node)

    def visit_Global(self, node: ast.Global):
        self.used_globals.update(node.names)
        self.generic_visit(node)

    def visit_Name(self, node: ast.Name):
        self.update_globals(node.id)
        self.generic_visit(node)

    def update_globals(self, name: str):
        if not self._check_assigned(name):
            self.used_globals.add(name)

    def _get_names(self, node):
        if isinstance(node, ast.Name):
            yield node.id
        elif isinstance(node, ast.Tuple):
            for n in node.elts:
                yield from self._get_names(n)
        elif isinstance(node, ast.Starred):
            yield from self._get_names(node.value)
        elif isinstance(node, ast.Call):
            yield from self._get_names(node.func)


def _get_funcdef_arg_names(node, attr) -> List[str]:
    arg_list = getattr(node.args, attr, None)
    if not arg_list:
        return []
    return list(map(lambda x: x.arg, getattr(node.args, attr)))



from types import FunctionType

from ..parse import parse_function, FuncDict
from ..__version import __version__

from .misc import get_diff_rc_params
from .metadata import get_metadata


def get_file_dict(func: FunctionType, *args, **kwargs) -> FuncDict:
    func_dict: FuncDict = parse_function(func)
    diff_rc = get_diff_rc_params()
    metadata = get_metadata()
    func_dict.update({
        'args': args,
        'kwargs': kwargs,
        '__version__': __version__,
        'metadata': metadata,
        'diff_rc': diff_rc,
    })
    return func_dict



from types import FunctionType
from typing import Union, List, Dict, Any, Tuple, Optional
from io import BytesIO
from pathlib import Path

from ..jupyter_support import create_new_jupyter_cell, display_code

from ..parse import FuncDict
from .save_read import save_file, read_file
from .file_dict import get_file_dict

__all__ = ['PltFile', 'save_plt_file', 'read_plt_file']


class PltFile(FuncDict):
    def get_code_str(self,
                     with_imports: bool = True,
                     with_subfunctions: bool = True,
                     with_globals: bool = True,
                     exec_line: bool = False,
                     comments: bool = True,
                     ):
        code: List[str] = []

        entry_func_name = self.entry_func
        entry_func = self.functions[entry_func_name]
        sub_funcs = [
            self.functions[func_name]
            for func_name in sorted(self.functions.keys())
            if func_name != entry_func_name
        ]

        if with_imports:
            import_lines = '\n'.join(sorted(self.import_lines))
            if comments:
                import_lines = f'# Imports:\n{import_lines}'
            code.append(import_lines)

        if with_globals:
            global_lines = '\n'.join([f'{k} = {repr(v)}' for k, v in self.global_vars.items()])
            if comments:
                global_lines = f'# Global variables:\n{global_lines}'
            code.append(global_lines)

        if comments:
            code.append(f'# Main function:')

        code.append(entry_func)

        if with_subfunctions:
            if comments:
                code.append('# Sub-functions:')
            code += sub_funcs

        if exec_line:
            exec_str = f'\n\n{self.entry_func}(*PltFile_ARGS, **PltFile_KWARGS)'
            if comments:
                exec_str = f'# Execute function:\n{exec_str}'
            code.append(exec_str)

        return '\n\n'.join(code)

    def exec(self):
        code = self.get_code_str(exec_line=True, comments=False)

        global_dict = {
            'PltFile_ARGS': self.args,
            'PltFile_KWARGS': self.kwargs,
        }
        exec(code, global_dict)

    @classmethod
    def from_func(cls, func: FunctionType, *args, **kwargs):
        func_dict = get_file_dict(func, *args, **kwargs)
        return cls(**func_dict)

    @classmethod
    def from_file(cls, file: Union[str, Path, BytesIO]):
        return cls(**read_file(file))

    def save(self, file: Union[str, Path, BytesIO]):
        save_file(file, self)

    def show_code(self, with_imports: bool = True, with_subfunctions: bool = True, with_globals: bool = True):
        code = self.get_code_str(with_imports, with_subfunctions, with_globals)
        try:
            create_new_jupyter_cell(code)
        except ImportError:
            return code

    def _ipython_display_(self):
        display_code(self.get_code_str())


def save_plt_file(
        file: Union[str, Path, BytesIO],
        func: FunctionType,
        args: Optional[Tuple[Any, ...]] = (),
        kwargs: Optional[Dict[str, Any]] = None,
        info_dict: Optional[Dict[str, Any]] = None,
) -> PltFile:
    plt_file = PltFile.from_func(func, *args, **kwargs)
    if info_dict:
        plt_file.update(info_dict)
    plt_file.save(file)
    return plt_file


def read_plt_file(file: Union[str, Path, BytesIO]) -> PltFile:
    return PltFile(**read_file(file))



from typing import Union

from pathlib import Path
from io import BytesIO

from .misc import _open_file_from_args, CorruptedPltFileError, NonSerializableDataError

from ..msgpack_import import msgpack
from ..utils import is_serializable, find_non_serializable


def read_file(file: Union[str, Path, BytesIO]) -> dict:
    with _open_file_from_args(file, 'rb') as f:
        try:
            return msgpack.unpack(f)
        except Exception as err:
            raise CorruptedPltFileError from err


def save_file(file: Union[str, Path, BytesIO], file_dict: dict):
    with _open_file_from_args(file, 'wb') as f:
        try:
            msgpack.pack(file_dict, f)
        except TypeError as err:
            keys = find_non_serializable(file_dict)

            internal_err_keys = []
            err_msg = f''
            raise NonSerializableDataError from err



from types import FunctionType
from typing import Tuple, Dict, Any, Optional
import datetime

from ..parse.parse_func_name import get_real_func_name
from ..utils import format_docs

__all__ = ['SaveOpts', 'add_opt_docs']

PLT_FORMAT: str = 'plt'

SAVE_OPTS_DOCSTRING: str = f"""
    name (str, optional): Name of the function. If not provided, __name__ attribute is used.
    rewrite (bool, optional): Rewrite the .{PLT_FORMAT} file. False by default.
    datefmt (str, optional): Date format for file name. Default format is '%d-%b-%H-%M-%S'.
        While providing a custom format, check the requirements of the OS regarding the allowed 
        charachters for file names.
    suffix (str, optional): Adds an suffix to the file name.
    save_figure (bool, optional): (experimental feature) Execute function and save the figure along 
        with the .{PLT_FORMAT} file. It is assumed that a function creates a figure via matplotlib package.
         False by default.
    fig_format (str, optional): (experimental feature) Figure format used if save_figure is True.
        Default format is 'eps'.
    save_plt (bool, optional): Save .{PLT_FORMAT} file. True by default. 
    description (str, optional): Description of the function to be saved in a .{PLT_FORMAT} file.
"""

add_opt_docs = format_docs(SAVE_OPTS_DOCSTRING, 'save_opt_args')


# dataclasses are not available in python 3.6 =/
class SaveOpts(dict):
    """
    Args:
        {save_opt_args}

    """

    _DEFAULTS = {
        'name': None,
        'rewrite': False,
        'suffix': None,
        'datefmt': '%d-%b-%H-%M-%S',
        'save_figure': False,
        'fig_format': 'eps',
        'save_plt': True,
        'description': None,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @property
    def info_dict(self) -> Dict[str, Any]:
        return {'description': self.description or ''}

    def new_opts(self, only_missing_keys: bool = False, **kwargs) -> 'SaveOpts':
        if only_missing_keys:
            base_kwargs, add_kwargs = kwargs, self
        else:
            base_kwargs, add_kwargs = self, kwargs

        opts_dict = SaveOpts(**base_kwargs)
        opts_dict.update(**add_kwargs)
        return opts_dict

    def __repr__(self):
        args = ', '.join([f'{k}={repr(v)}' for k, v in self.items()])
        return f'SaveOpts({args})'

    @property
    def time_str(self) -> str:
        return datetime.datetime.now().strftime(self.datefmt)

    def get_names(self, func: FunctionType = None) -> Tuple[str, str]:
        func_name = self._get_func_name(func)

        return self._get_file_names_from_func_name(func_name)

    def _fget(self, name: str):
        return self.get(name, self._DEFAULTS[name])

    @property
    def name(self):
        return self._fget('name')

    @property
    def rewrite(self):
        return self._fget('rewrite')

    @property
    def suffix(self):
        return self._fget('suffix')

    @property
    def datefmt(self):
        return self._fget('datefmt')

    @property
    def save_figure(self):
        return self._fget('save_figure')

    @property
    def fig_format(self):
        return self._fget('fig_format')

    @property
    def save_plt(self):
        return self._fget('save_plt')

    @property
    def description(self):
        return self._fget('description')

    def _get_func_name(self, func: FunctionType = None) -> str:
        func_name = self.name or _get_func_name(func)

        if self.suffix:
            func_name = f'{func_name}_{self.suffix}'

        if not self.rewrite:
            func_name = f'{func_name}_{self.time_str}'

        return func_name

    def _get_file_names_from_func_name(self, func_name: str) -> Tuple[str, str]:
        return f'{func_name}.{PLT_FORMAT}', f'{func_name}.{self.fig_format}'


def _get_func_name(func: FunctionType) -> str:
    try:
        return func.__name__
    except AttributeError:
        return get_real_func_name(func)



import unicodedata
import re
from typing import Optional, Any
from types import FunctionType

from .msgpack_import import msgpack


__all__ = ['find_non_serializable', 'is_serializable', 'str2valid_filename']


def format_docs(docs: str, docs_name: Optional[str] = None) -> object:
    """
    Returns a decorator that modifies an object docstring by either inserting docs instead of '{[docs_name]}'
    ([docs_name] stands for docs_name keyword here) or by replacing docstring by docs if docs_name
    is not provided.

    Args:
        docs (str): Docstring to be inserted (or replaced) into the decorated object docstring.
        docs_name (str, optional): Format keyword.

    Returns:
        FunctionType: A decorator that modifies object docstring and returns the object.

    Examples:
        >>> @format_docs('Some docstring.')
        ... def func1():
        ...     pass
        ...
        >>> func1.__doc__
        'Some docstring.'

        >>> @format_docs('arguments', 'docs')
        ... def func2():
        ...     '''Docstring with {docs}.'''
        ...     pass
        ...
        >>> func2.__doc__
        'Docstring with arguments.'

        >>> @format_docs('arguments', 'docs')
        ... def func2():
        ...     '''Docstring with {docs}.'''
        ...     pass
        ...
        >>> func2.__doc__
        'Docstring with arguments.'

        >>> docs = '''
        ... argument1
        ... argument2'''

        >>> @format_docs(docs, 'docs')
        ... def func3():
        ...     '''Multiline docstring.
        ...         {docs}
        ... '''
        ...     pass
        ...
        >>> print(func3.__doc__)
        Multiline docstring.
        <BLANKLINE>
                argument1
                argument2
        <BLANKLINE>
        >>> @format_docs(docs, 'docs')
        ... def func4():
        ...     '''
        ...     Multiline docstring.
        ...         {docs}
        ... '''
        ...     pass
        ...
        >>> print(func4.__doc__)
        <BLANKLINE>
            Multiline docstring.
        <BLANKLINE>
            argument1
            argument2
        <BLANKLINE>
    """

    def decorator(obj: object) -> object:
        if docs_name:
            obj.__doc__ = _modify_docstring(obj.__doc__, docs, docs_name)
            return obj
        else:
            obj.__doc__ = docs
        return obj

    return decorator


def _modify_docstring(src_doc: str, new_doc: str, docs_name: str) -> str:
    key = '{%s}' % docs_name
    doc_lines = src_doc.split('\n')
    first_doc_line = ''

    for num, line in enumerate(doc_lines):
        if not first_doc_line and line:
            first_doc_line = line
        if key in line:
            break
    else:  # no docs_name found
        return src_doc

    init_indentation: int = _get_indentation(first_doc_line)

    target_indentation: int = _get_indentation(line.split(key)[0]) - init_indentation

    indentation = ' ' * target_indentation

    formatted_docs = f'\n{indentation}'.join(new_doc.split('\n'))

    doc_lines[num] = line.replace(key, formatted_docs)

    return '\n'.join(doc_lines)


def _get_indentation(src: str) -> int:
    src = src.replace('\t', ' ' * 4)
    return len(src) - len(src.lstrip(' '))


def str2valid_filename(value: str) -> str:
    """
    Adapted from https://github.com/django/django/blob/master/django/utils/text.py
    Convert spaces or repeated dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Convert to lowercase. Also strip leading and
    trailing whitespace, dashes, and underscores.

    Args:
        value (str): Initial name of a file.
    Returns:
        str: A valid file name.

    Examples:
        >>> str2valid_filename('2 Jun 02:15:45')
        '2-jun-02-15-45'

        >>> str2valid_filename('path/to\dir')
        'path-to-dir'
    """
    value = unicodedata.normalize('NFKC', value)
    value = re.sub(r'[^\w\s-]', ' ', value.lower())
    return re.sub(r'[-\s]+', '-', value).strip('-_')


def is_serializable(data) -> bool:
    try:
        msgpack.packb(data)
        return True
    except TypeError:
        return False


def find_non_serializable(data: dict or list):
    return _find_non_serializable(data)


def _find_non_serializable(data: dict or list, *, prev_keys=None) -> list:
    _found_keys = []

    prev_keys = prev_keys or []

    if isinstance(data, (list, tuple)):
        data = {str(i): v for i, v in enumerate(data)}

    for key, value in data.items():
        if isinstance(value, (list, tuple, dict)):
            _found_keys += _find_non_serializable(value, prev_keys=prev_keys + [key])
        else:
            if not is_serializable(value):
                _found_keys.append(prev_keys + [key])
    return _found_keys



from typing import Dict, Tuple, Any

__all__ = ['FuncDict']


class FuncDict(dict):
    def __init__(self,
                 entry_func: str,
                 functions: Dict[str, str],
                 modules: Tuple[str, ...],
                 import_lines: Tuple[str, ...],
                 module_versions: Dict[str, str],
                 global_vars: Dict[str, Any],
                 **kwargs
                 ):
        super().__init__()
        self.update(dict(
            entry_func=entry_func,
            functions=functions,
            modules=modules,
            import_lines=import_lines,
            module_versions=module_versions,
            global_vars=global_vars,
        ))
        self.update(kwargs)

    # Mandatory keys
    @property
    def entry_func(self) -> str:
        return self['entry_func']

    @property
    def functions(self) -> Dict[str, str]:
        return self['functions']

    @property
    def modules(self) -> Tuple[str, ...]:
        return self['modules']

    @property
    def import_lines(self) -> Tuple[str, ...]:
        return self['import_lines']

    @property
    def module_versions(self) -> Dict[str, str]:
        return self['module_versions']

    @property
    def global_vars(self) -> Dict[str, Any]:
        return self['global_vars']

    # Optional keys

    @property
    def args(self) -> Tuple[Any, ...]:
        return self.get('args', ())

    @property
    def kwargs(self) -> Dict[str, Any]:
        return self.get('kwargs', {})

    @property
    def version(self) -> str:
        return self.get('__version__', '')

    @property
    def metadata(self) -> str:
        return self.get('metadata', '')

    @property
    def diff_rc(self) -> Dict[str, Any]:
        return self.get('diff_rc', {})

    @property
    def description(self) -> str:
        return self.get('description', '')
